@@ -33,11 +33,33 @@
 #include <string.h>
 #include <errno.h>
 #include <unistd.h>
-#ifdef __linux__
-#include <termio.h>
-#else
 #include <termios.h>
-#endif /* of __linux__ */
+
+/* glibc for MIPS has its own bits/termios.h which does not define
+ * CMSPAR, so we use the value from the generic bits/termios.h
+ */
+#ifdef __linux__
+#ifndef CMSPAR
+#define CMSPAR 010000000000
+#endif
+#endif
+
+/* Some BSDs (and possibly other systems too) have no mark / space
+ * parity support, and they don't define CMSPAR. Use a zero CMSPAR in
+ * these cases. If the user tries to set P_MARK or P_SPACE he will get
+ * P_EVEN or P_ODD instead. */
+#ifndef CMSPAR
+#define CMSPAR 0
+#endif
+
+#ifdef __linux__
+#include <sys/ioctl.h>
+#endif
+
+#ifdef USE_CUSTOM_BAUD
+/* only works for linux, recent kernels */
+#include "termios2.h"
+#endif
 
 #include "term.h"
 
@@ -69,11 +91,14 @@ static const char * const term_err_str[] = {
 	[TERM_EBAUD]      = "Invalid baud rate",
 	[TERM_ESETOSPEED] = "Cannot set the output speed",
 	[TERM_ESETISPEED] = "Cannot set the input speed",
+	[TERM_EGETSPEED]  = "Cannot decode speed",
 	[TERM_EPARITY]    = "Invalid parity mode",
 	[TERM_EDATABITS]  = "Invalid number of databits",
+	[TERM_ESTOPBITS]  = "Invalid number of stopbits",
 	[TERM_EFLOW]      = "Invalid flowcontrol mode",
     [TERM_EDTRDOWN]   = "Cannot lower DTR",
     [TERM_EDTRUP]     = "Cannot raise DTR",
+	[TERM_EMCTL]      = "Cannot get mctl status",
 	[TERM_EDRAIN]     = "Cannot drain the device",
 	[TERM_EBREAK]     = "Cannot send break sequence"
 };
@@ -107,9 +132,11 @@ term_strerror (int terrnum, int errnum)
 	case TERM_EBAUD:
 	case TERM_EPARITY:
 	case TERM_EDATABITS:
+	case TERM_ESTOPBITS:
 	case TERM_EFLOW:
 	case TERM_EDTRDOWN:
 	case TERM_EDTRUP:
+	case TERM_EMCTL:
 		snprintf(term_err_buff, sizeof(term_err_buff),
 				 "%s", term_err_str[terrnum]);
 		rval = term_err_buff;
@@ -131,6 +158,150 @@ term_perror (const char *prefix)
 
 /***************************************************************************/
 
+#define BNONE 0xFFFFFFFF
+
+struct baud_codes {
+	int speed;
+	speed_t code;
+} baud_table[] = {
+	{ 0, B0 },
+	{ 50, B50 },
+	{ 75, B75 },
+	{ 110, B110 },
+	{ 134, B134 },
+	{ 150, B150 },
+	{ 200, B200 },
+	{ 300, B300 },
+	{ 600, B600 },
+	{ 1200, B1200 },
+	{ 1800, B1800 },
+	{ 2400, B2400 },
+	{ 4800, B4800 },
+	{ 9600, B9600 },
+	{ 19200, B19200 },
+	{ 38400, B38400 },
+	{ 57600, B57600 },
+	{ 115200, B115200 },
+#ifdef HIGH_BAUD
+#ifdef B230400
+	{ 230400, B230400 },
+#endif
+#ifdef B460800
+	{ 460800, B460800 },
+#endif
+#ifdef B500000
+	{ 500000, B500000 },
+#endif
+#ifdef B576000
+	{ 576000, B576000 },
+#endif
+#ifdef B921600
+	{ 921600, B921600 },
+#endif
+#ifdef B1000000
+	{ 1000000, B1000000 },
+#endif
+#ifdef B1152000
+	{ 1152000, B1152000 },
+#endif
+#ifdef B1500000
+	{ 1500000, B1500000 },
+#endif
+#ifdef B2000000
+	{ 2000000, B2000000 },
+#endif
+#ifdef B2500000
+	{ 2500000, B2500000 },
+#endif
+#ifdef B3000000
+	{ 3000000, B3000000 },
+#endif
+#ifdef B3500000
+	{ 3500000, B3500000 },
+#endif
+#ifdef B4000000
+	{ 4000000, B4000000 },
+#endif
+#endif /* of HIGH_BAUD */
+};
+
+#define BAUD_TABLE_SZ (sizeof(baud_table) / sizeof(baud_table[0]))
+
+int
+term_baud_up (int baud)
+{
+	int i;
+
+	for (i = 0; i < BAUD_TABLE_SZ; i++) {
+		if ( baud >= baud_table[i].speed )
+			continue;
+		else {
+			baud = baud_table[i].speed;
+			break;
+		}
+	}
+
+ 	return baud;
+}
+
+int
+term_baud_down (int baud)
+{
+	int i;
+
+	for (i = BAUD_TABLE_SZ - 1; i >= 0; i--) {
+		if ( baud <= baud_table[i].speed )
+			continue;
+		else {
+			baud = baud_table[i].speed;
+			break;
+		}
+	}
+
+	return baud;
+}
+
+static speed_t
+Bcode(int speed)
+{
+	speed_t code = BNONE;
+	int i;
+
+	for (i = 0; i < BAUD_TABLE_SZ; i++) {
+		if ( baud_table[i].speed == speed ) {
+			code = baud_table[i].code;
+			break;
+		}
+	}
+	return code;
+}
+
+static int
+Bspeed(speed_t code)
+{
+	int speed = -1, i;
+
+	for (i = 0; i < BAUD_TABLE_SZ; i++) {
+		if ( baud_table[i].code == code ) {
+			speed = baud_table[i].speed;
+			break;
+		}
+	}
+	return speed;
+}
+
+int
+term_baud_ok(int baud)
+{
+#ifndef USE_CUSTOM_BAUD
+	return (Bcode(baud) != BNONE) ? 1 : 0;
+#else
+	return (baud >= 0);
+#endif
+}
+
+/**************************************************************************/
+
 static int
 term_find_next_free (void)
 {
@@ -201,12 +372,10 @@ term_exitfunc (void)
 		for (i = 0; i < MAX_TERMS; i++) {
 			if (term.fd[i] == -1)
 				continue;
-			do { /* dummy */
-				r = tcflush(term.fd[i], TCIOFLUSH);
-				if ( r < 0 ) break;
-				r = tcsetattr(term.fd[i], TCSAFLUSH, &term.origtermios[i]);
-				if ( r < 0 ) break;
-			} while (0);
+			tcflush(term.fd[i], TCIOFLUSH);
+			do {
+				r = tcsetattr(term.fd[i], TCSANOW, &term.origtermios[i]);
+			} while ( r < 0 && errno == EINTR );
 			if ( r < 0 ) {
 				char *tname;
 
@@ -235,12 +404,10 @@ term_lib_init (void)
 			for (i = 0; i < MAX_TERMS; i++) {
 				if (term.fd[i] == -1)
 					continue;
+				tcflush(term.fd[i], TCIOFLUSH);
 				do {
-					r = tcflush(term.fd[i], TCIOFLUSH);
-					if ( r < 0 ) break;
-					r = tcsetattr(term.fd[i], TCSAFLUSH, &term.origtermios[i]);
-					if ( r < 0 ) break;
-				} while (0);
+					r = tcsetattr(term.fd[i], TCSANOW, &term.origtermios[i]);
+				} while ( r < 0 && errno == EINTR );
 				if ( r < 0 ) {
 					char *tname;
  
@@ -335,7 +502,7 @@ term_remove(int fd)
 				rval = -1;
 				break;
 			}
-			r = tcsetattr(term.fd[i], TCSAFLUSH, &term.origtermios[i]);
+			r = tcsetattr(term.fd[i], TCSANOW, &term.origtermios[i]);
 			if ( r < 0 ) {
 				term_errno = TERM_ESETATTR;
 				rval = -1;
@@ -388,12 +555,18 @@ term_replace (int oldfd, int newfd)
 			break;
 		}
 
-		r = tcsetattr(newfd, TCSAFLUSH, &term.currtermios[i]);
+		r = tcsetattr(newfd, TCSANOW, &term.currtermios[i]);
 		if ( r < 0 ) {
 			term_errno = TERM_ESETATTR;
 			rval = -1;
 			break;
 		}
+		r = tcgetattr(newfd, &term.currtermios[i]);
+		if ( r < 0 ) {
+			term_errno = TERM_EGETATTR;
+			rval = -1;
+			break;
+		}
 
 		term.fd[i] = newfd;
 
@@ -425,15 +598,20 @@ term_reset (int fd)
 			rval = -1;
 			break;
 		}
-		r = tcsetattr(term.fd[i], TCSAFLUSH, &term.origtermios[i]);
+		r = tcsetattr(term.fd[i], TCSANOW, &term.origtermios[i]);
 		if ( r < 0 ) {
 			term_errno = TERM_ESETATTR;
 			rval = -1;
 			break;
 		}
+		r = tcgetattr(term.fd[i], &term.currtermios[i]);
+		if ( r < 0 ) {
+			term_errno = TERM_EGETATTR;
+			rval = -1;
+			break;
+		}
 
-		term.currtermios[i] = term.origtermios[i];
-		term.nexttermios[i] = term.origtermios[i];
+		term.nexttermios[i] = term.currtermios[i];
 	} while (0);
 
 	return rval;
@@ -495,9 +673,11 @@ term_refresh (int fd)
 /***************************************************************************/
 
 int
-term_apply (int fd)
+term_apply (int fd, int now)
 {
-	int rval, r, i;
+	int when, rval, r, i;
+
+	when = now ? TCSANOW : TCSAFLUSH;
 
 	rval = 0;
 
@@ -509,13 +689,19 @@ term_apply (int fd)
 			break;
 		}
 		
-		r = tcsetattr(term.fd[i], TCSAFLUSH, &term.nexttermios[i]);
+		r = tcsetattr(term.fd[i], when, &term.nexttermios[i]);
 		if ( r < 0 ) {
 			term_errno = TERM_ESETATTR;
 			rval = -1;
 			break;
 		}
-		
+		r = tcgetattr(term.fd[i], &term.nexttermios[i]);
+		if ( r < 0 ) {
+			term_errno = TERM_EGETATTR;
+			rval = -1;
+			break;
+		}
+
 		term.currtermios[i] = term.nexttermios[i];
 
 	} while (0);
@@ -571,92 +757,29 @@ term_set_baudrate (int fd, int baudrate)
 		}
 
 		tio = term.nexttermios[i];
-
-		switch (baudrate) {
-		case 0:
-			spd = B0;
-			break;
-		case 50:
-			spd = B50;
-			break;
-		case 75:
-			spd = B75;
-			break;
-		case 110:
-			spd = B110;
-			break;
-		case 134:
-			spd = B134;
-			break;
-		case 150:
-			spd = B150;
-			break;
-		case 200:
-			spd = B200;
-			break;
-		case 300:
-			spd = B300;
-			break;
-		case 600:
-			spd = B600;
-			break;
-		case 1200:
-			spd = B1200;
-			break;
-		case 1800:
-			spd = B1800;
-			break;
-		case 2400:
-			spd = B2400;
-			break;
-		case 4800:
-			spd = B4800;
-			break;
-		case 9600:
-			spd = B9600;
-			break;
-		case 19200:
-			spd = B19200;
-			break;
-		case 38400:
-			spd = B38400;
-			break;
-		case 57600:
-			spd = B57600;
-			break;
-		case 115200:
-			spd = B115200;
-			break;
-#ifdef HIGH_BAUD
-		case 230400:
-			spd = B230400;
-			break;
-		case 460800:
-			spd = B460800;
-			break;
-		case 921600:
-			spd = B921600;
-			break;
-#endif
-		default:
+		spd = Bcode(baudrate);
+		if ( spd != BNONE ) {
+			r = cfsetospeed(&tio, spd);
+			if ( r < 0 ) {
+				term_errno = TERM_ESETOSPEED;
+				rval = -1;
+				break;
+			}
+			cfsetispeed(&tio, B0);
+		} else {
+#ifdef USE_CUSTOM_BAUD
+			r = cfsetospeed_custom(&tio, baudrate);
+			if ( r < 0 ) {
+				term_errno = TERM_ESETOSPEED;
+				rval = -1;
+				break;
+			}
+			cfsetispeed(&tio, B0);
+#else /* ! defined USE_CUSTOM_BAUD */
 			term_errno = TERM_EBAUD;
 			rval = -1;
 			break;
-		}
-		if ( rval < 0 ) break;
-
-		r = cfsetospeed(&tio, spd);
-		if ( r < 0 ) {
-			term_errno = TERM_ESETOSPEED;
-			rval = -1;
-			break;
-		}
-			
-		r = cfsetispeed(&tio, spd);
-		if ( r < 0 ) {
-			term_errno = TERM_ESETISPEED;
-			rval = -1;
-			break;
+#endif /* of USE_CUSTOM_BAUD */
 		}
 
 		term.nexttermios[i] = tio;
@@ -666,6 +789,47 @@ term_set_baudrate (int fd, int baudrate)
 	return rval;
 }
 
+int 
+term_get_baudrate (int fd, int *ispeed)
+{
+	speed_t code;
+	int i, ospeed;
+
+	do { /* dummy */
+
+		i = term_find(fd);
+		if ( i < 0 ) {
+			ospeed = -1;
+			break;
+		}
+
+		if ( ispeed ) {
+			code = cfgetispeed(&term.currtermios[i]);
+			*ispeed = Bspeed(code);
+#ifdef USE_CUSTOM_BAUD
+			if ( *ispeed < 0 ) {
+				*ispeed = cfgetispeed_custom(&term.currtermios[i]);
+			}
+#endif
+		}
+		code = cfgetospeed(&term.currtermios[i]);
+		ospeed = Bspeed(code);
+		if ( ospeed < 0 ) {
+#ifdef USE_CUSTOM_BAUD
+			ospeed = cfgetospeed_custom(&term.currtermios[i]);
+			if ( ospeed < 0 ) {
+				term_errno = TERM_EGETSPEED;
+			}
+#else
+			term_errno = TERM_EGETSPEED;
+#endif
+		}
+
+	} while (0);
+
+	return ospeed;
+}
+
 /***************************************************************************/
 
 int
@@ -688,14 +852,22 @@ term_set_parity (int fd, enum parity_e parity)
 
 		switch (parity) {
 		case P_EVEN:
-			tiop->c_cflag &= ~PARODD;
+			tiop->c_cflag &= ~(PARODD | CMSPAR);
 			tiop->c_cflag |= PARENB;
 			break;
 		case P_ODD:
+			tiop->c_cflag &= ~CMSPAR;
 			tiop->c_cflag |= PARENB | PARODD;
 			break;
+		case P_MARK:
+			tiop->c_cflag |= PARENB | PARODD | CMSPAR;
+			break;
+		case P_SPACE:
+			tiop->c_cflag &= ~PARODD;
+			tiop->c_cflag |= PARENB | CMSPAR;
+			break;
 		case P_NONE:
-			tiop->c_cflag &= ~(PARENB | PARODD); 
+			tiop->c_cflag &= ~(PARENB | PARODD | CMSPAR); 
 			break;
 		default:
 			term_errno = TERM_EPARITY;
@@ -709,6 +881,34 @@ term_set_parity (int fd, enum parity_e parity)
 	return rval;
 }
 
+enum parity_e
+term_get_parity (int fd)
+{
+	tcflag_t flg;
+	int i, parity;
+
+	do { /* dummy */
+
+		i = term_find(fd);
+		if ( i < 0 ) {
+			parity = -1;
+			break;
+		}
+
+		flg = term.currtermios[i].c_cflag;
+		if ( ! (flg & PARENB) ) {
+			parity = P_NONE;
+		} else if ( flg & CMSPAR ) {
+			parity = (flg & PARODD) ? P_MARK : P_SPACE;
+		} else {
+			parity = (flg & PARODD) ? P_ODD : P_EVEN;
+		}
+
+	} while (0);
+	
+	return parity;
+}
+
 /***************************************************************************/
 
 int
@@ -754,6 +954,101 @@ term_set_databits (int fd, int databits)
 	return rval;
 }
 
+int
+term_get_databits (int fd)
+{
+	tcflag_t flg;
+	int i, bits;
+
+	do { /* dummy */
+
+		i = term_find(fd);
+		if ( i < 0 ) {
+			bits = -1;
+			break;
+		}
+
+		flg = term.currtermios[i].c_cflag & CSIZE;
+		switch (flg) {
+		case CS5:
+			bits = 5;
+			break;
+		case CS6:
+			bits = 6;
+			break;
+		case CS7:
+			bits = 7;
+			break;
+		case CS8:
+		default:
+			bits = 8;
+			break;
+		}
+
+	} while (0);
+
+	return bits;
+}
+
+/***************************************************************************/
+
+int
+term_set_stopbits (int fd, int stopbits)
+{
+	int rval, i;
+	struct termios *tiop;
+
+	rval = 0;
+
+	do { /* dummy */
+
+		i = term_find(fd);
+		if ( i < 0 ) {
+			rval = -1;
+			break;
+		}
+
+		tiop = &term.nexttermios[i];
+				
+		switch (stopbits) {
+		case 1:
+			tiop->c_cflag &= ~CSTOPB;
+			break;
+		case 2:
+			tiop->c_cflag |= CSTOPB;
+			break;
+		default:
+			term_errno = TERM_ESTOPBITS;
+			rval = -1;
+			break;
+		}
+		if ( rval < 0 ) break;
+
+	} while (0);
+
+	return rval;
+}
+
+int
+term_get_stopbits (int fd)
+{
+	int i, bits;
+
+	do { /* dummy */
+
+		i = term_find(fd);
+		if ( i < 0 ) {
+			bits = -1;
+			break;
+		}
+
+		bits = (term.currtermios[i].c_cflag & CSTOPB) ? 2 : 1;
+
+	} while (0);
+
+	return bits;
+}
+
 /***************************************************************************/
 
 int
@@ -799,6 +1094,39 @@ term_set_flowcntrl (int fd, enum flowcntrl_e flowcntl)
 	return rval;
 }
 
+enum flowcntrl_e
+term_get_flowcntrl (int fd)
+{
+	int i, flow;
+	int rtscts, xoff, xon;
+
+	do { /* dummy */
+
+		i = term_find(fd);
+		if ( i < 0 ) {
+			flow = -1;
+			break;
+		}
+
+		rtscts = (term.currtermios[i].c_cflag & CRTSCTS) ? 1 : 0;
+		xoff = (term.currtermios[i].c_iflag & IXOFF) ? 1 : 0;
+		xon = (term.currtermios[i].c_iflag & (IXON | IXANY)) ? 1 : 0;
+
+		if ( rtscts && ! xoff && ! xon ) {
+			flow = FC_RTSCTS;
+		} else if ( ! rtscts && xoff && xon ) {
+			flow = FC_XONXOFF;
+		} else if ( ! rtscts && ! xoff && ! xon ) {
+			flow = FC_NONE;
+		} else {
+			flow = FC_OTHER;
+		}
+
+	} while (0);
+	
+	return flow;
+}
+
 /***************************************************************************/
 
 int
@@ -864,7 +1192,10 @@ term_set_hupcl (int fd, int on)
 int
 term_set(int fd,
 		 int raw,
-		 int baud, enum parity_e parity, int bits, enum flowcntrl_e fc,
+		 int baud, 
+		 enum parity_e parity, 
+		 int databits, int stopbits,
+		 enum flowcntrl_e fc,
 		 int local, int hup_close)
 {
 	int rval, r, i, ni;
@@ -900,7 +1231,10 @@ term_set(int fd,
 			r = term_set_parity(fd, parity);
 			if ( r < 0 ) { rval = -1; break; }
 			
-			r = term_set_databits(fd, bits);
+			r = term_set_databits(fd, databits);
+			if ( r < 0 ) { rval = -1; break; }
+
+			r = term_set_stopbits(fd, stopbits);
 			if ( r < 0 ) { rval = -1; break; }
 			
 			r = term_set_flowcntrl(fd, fc);
@@ -1105,6 +1439,44 @@ term_lower_dtr(int fd)
 
 /***************************************************************************/
 
+int
+term_get_mctl (int fd)
+{
+	int mctl, i;
+
+	do { /* dummy */
+
+		i = term_find(fd);
+		if ( i < 0 ) { 
+			mctl = -1;
+			break;
+		}
+
+#ifdef __linux__
+		{ 
+			int r, pmctl;
+			
+			r = ioctl(fd, TIOCMGET, &pmctl);
+			if (r < 0) {
+				mctl = -1;
+				break;
+			}
+			mctl = 0;
+			if (pmctl & TIOCM_DTR) mctl |= MCTL_DTR;
+			if (pmctl & TIOCM_DSR) mctl |= MCTL_DSR;
+			if (pmctl & TIOCM_CD) mctl |= MCTL_DCD;
+			if (pmctl & TIOCM_RTS) mctl |= MCTL_RTS;
+			if (pmctl & TIOCM_CTS) mctl |= MCTL_CTS;
+			if (pmctl & TIOCM_RI) mctl |= MCTL_RI;
+		}
+#else
+		mctl = MCTL_UNAVAIL;
+#endif
+	} while(0);
+
+	return mctl;
+}
+
 int
 term_drain(int fd)
 {
@@ -1121,7 +1493,12 @@ term_drain(int fd)
 		}
 
 		do {
+#ifdef __BIONIC__
+			/* See: http://dan.drown.org/android/src/gdb/no-tcdrain */
+			r = ioctl(fd, TCSBRK, 1);
+#else
 			r = tcdrain(fd);
+#endif
 		} while ( r < 0 && errno == EINTR);
 		if ( r < 0 ) {
 			term_errno = TERM_EDRAIN;
